@@ -120,19 +120,23 @@ udp_recv(Sock, Addr, Port, Data, State) ->
 %%====================================================================
 init([Sock, Opts]) ->
     process_flag(trap_exit, true),
-    case inet:peername(Sock) of
+    case get_peername(Sock, Opts) of
 	{ok, Addr} ->
-	    TRef = erlang:start_timer(?TIMEOUT, self(), stop),
-	    SockMod = get_sockmod(Opts),
-	    State = prepare_state(Opts, Sock, Addr, SockMod),
-	    CertFile = get_certfile(Opts),
-	    case maybe_starttls(Sock, SockMod, CertFile) of
-		{ok, NewSock} ->
-		    inet:setopts(Sock, [{active, once}]),
-		    {ok, session_established,
-		     State#state{tref = TRef, sock = NewSock}};
-		{error, Why} ->
-		    {stop, Why}
+	    case get_sockmod(Opts, Sock) of
+		{ok, SockMod} ->
+		    State = prepare_state(Opts, Sock, Addr, SockMod),
+		    CertFile = get_certfile(Opts),
+		    case maybe_starttls(Sock, SockMod, CertFile) of
+			{ok, NewSock} ->
+			    inet:setopts(Sock, [{active, once}]),
+			    TRef = erlang:start_timer(?TIMEOUT, self(), stop),
+			    {ok, session_established,
+			     State#state{tref = TRef, sock = NewSock}};
+			{error, Reason} ->
+			    {stop, Reason}
+		    end;
+		{error, Reason} ->
+		    {stop, Reason}
 	    end;
 	Err ->
 	    {stop, Err}
@@ -240,9 +244,8 @@ process(#state{auth = user} = State,
 		    ?LOG_NOTICE("Failed long-term STUN/TURN authentication"),
 		    send(NewState, R);
 		Pass ->
-		    Key = {User, Realm, Pass},
-		    case stun_codec:check_integrity(Msg, Key) of
-			true ->
+		    case check_integrity(User, Realm, Msg, Pass) of
+			{true, Key} ->
 			    ?LOG_INFO("Accepting long-term STUN/TURN "
 				      "authentication"),
 			    process(NewState, Msg, Key);
@@ -571,6 +574,7 @@ prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
 		 ({certfile, _}, State) -> State;
 		 ({tls, _}, State) -> State;
 		 (tls, State) -> State;
+		 ({sock_peer_name, _}, State) -> State;
 		 ({session_id, _}, State) -> State;
 		 (Opt, State) ->
 		      ?LOG_ERROR("Ignoring unknown option '~p'", [Opt]),
@@ -656,6 +660,19 @@ have_nonce(Nonce, Nonces) ->
 	    {false, NewNonces}
     end.
 
+check_integrity(User, Realm, Msg, Pass) when is_binary(Pass) ->
+    check_integrity(User, Realm, Msg, [Pass]);
+check_integrity(_User, _Realm, _Msg, []) ->
+    false;
+check_integrity(User, Realm, Msg, [Pass | T]) ->
+    Key = {User, Realm, Pass},
+    case stun_codec:check_integrity(Msg, Key) of
+	true ->
+	    {true, Key};
+	false ->
+	    check_integrity(User, Realm, Msg, T)
+    end.
+
 unmap_v4_addr({{0, 0, 0, 0, 0, 16#FFFF, D7, D8}, Port}) ->
     {{D7 bsr 8, D7 band 255, D8 bsr 8, D8 band 255}, Port};
 unmap_v4_addr(AddrPort) ->
@@ -680,12 +697,29 @@ is_valid_subnet({{IP1, IP2, IP3, IP4, IP5, IP6, IP7, IP8}, Mask}) ->
 is_valid_subnet(_) ->
     false.
 
-get_sockmod(Opts) ->
-    case proplists:get_bool(tls, Opts) of
+get_sockmod(Opts, Sock) ->
+    case proplists:get_value(tls, Opts, false) of
 	true ->
-	    fast_tls;
+	    {ok, fast_tls};
 	false ->
-	    gen_tcp
+	    {ok, gen_tcp};
+	optional ->
+	    case is_tls_handshake(Sock) of
+		true ->
+		    {ok, fast_tls};
+		false ->
+		    {ok, gen_tcp};
+		{error, _Reason} = Err ->
+		    Err
+	    end
+    end.
+
+get_peername(Sock, Opts) ->
+    case proplists:get_value(sock_peer_name, Opts) of
+	{_, Addr} ->
+	    {ok, Addr};
+	undefined ->
+	    inet:peername(Sock)
     end.
 
 get_certfile(Opts) ->
@@ -696,6 +730,26 @@ get_certfile(Opts) ->
 	    undefined
     end.
 
+-ifdef(USE_OLD_INET_BACKEND).
+-dialyzer({[no_match], [get_sockmod/2]}).
+is_tls_handshake(_Sock) ->
+    ?LOG_ERROR("Multiplexing TCP and TLS requires a newer Erlang/OTP version"),
+    {error, eprotonosupport}.
+-else.
+is_tls_handshake({_, _, {_, Socket}}) ->
+    case socket:recvfrom(Socket, 10, [peek], ?TIMEOUT) of
+	{ok, {_, <<22, 3, _:4/binary, 0, _:2/binary, 3>>}} ->
+	    ?LOG_DEBUG("Determined transport protocol: TLS"),
+	    true;
+	{ok, {_, _}} ->
+	    ?LOG_DEBUG("Determined transport protocol: TCP"),
+	    false;
+	{error, Reason} = Err ->
+	    ?LOG_INFO("Cannot determine transport protocol: ~s", [Reason]),
+	    Err
+    end.
+-endif.
+
 maybe_starttls(_Sock, fast_tls, undefined) ->
     ?LOG_ERROR("Cannot start TLS connection: option 'certfile' is not set"),
     {error, eprotonosupport};
