@@ -30,11 +30,13 @@
 
 %% gen_server callbacks
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
-         terminate/2, code_change/3]).
+	 terminate/2, code_change/3]).
 
 -include("stun_logger.hrl").
 
 -define(TCP_SEND_TIMEOUT, 10000).
+-define(UDP_READ_PACKETS, 100).
+-define(UDP_RECBUF, 1024 * 1024). % 1 MiB
 -record(state, {listeners = #{}}).
 
 %%%===================================================================
@@ -124,32 +126,46 @@ code_change(_OldVsn, State, _Extra) ->
 %%%===================================================================
 %%% Internal functions
 %%%===================================================================
-start_listener(IP, Port, Transport, Opts, Owner)
-  when Transport == tcp; Transport == tls ->
-    OptsWithTLS = case Transport of
-		      tls -> [tls|Opts];
-		      tcp -> Opts
-		  end,
-    case gen_tcp:listen(Port, [binary,
-                               {ip, IP},
-                               {packet, 0},
-                               {active, false},
-                               {reuseaddr, true},
-                               {nodelay, true},
-                               {keepalive, true},
-			       {send_timeout, ?TCP_SEND_TIMEOUT},
-			       {send_timeout_close, true}]) of
-        {ok, ListenSocket} ->
-            Owner ! {self(), ok},
-	    OptsWithTLS1 = stun:tcp_init(ListenSocket, OptsWithTLS),
-            accept(ListenSocket, OptsWithTLS1);
-        Err ->
-            Owner ! {self(), Err}
+start_listener(IP, Port, Transport0, Opts0, Owner)
+  when Transport0 == tcp;
+       Transport0 == tls;
+       Transport0 == auto ->
+    {Transport, Opts} = case {Transport0, proplists:get_value(tls, Opts0)} of
+			    {tcp, false} ->
+				{tcp, Opts0};
+			    {tcp, true} ->
+				{tls, Opts0};
+			    {tcp, optional} ->
+				{auto, Opts0};
+			    {tls, undefined} ->
+				{tls, [tls | Opts0]};
+			    {auto, undefined} ->
+				{auto, [{tls, optional} | Opts0]};
+			    {_Transport, _TLS} ->
+				{Transport0, Opts0}
+			end,
+    case listen(Transport, Port, [binary,
+				  {ip, IP},
+				  {packet, 0},
+				  {active, false},
+				  {reuseaddr, true},
+				  {nodelay, true},
+				  {keepalive, true},
+				  {send_timeout, ?TCP_SEND_TIMEOUT},
+				  {send_timeout_close, true}]) of
+	{ok, ListenSocket} ->
+	    Owner ! {self(), ok},
+	    Opts1 = stun:tcp_init(ListenSocket, Opts),
+	    accept(Transport, ListenSocket, Opts1);
+	Err ->
+	    Owner ! {self(), Err}
     end;
 start_listener(IP, Port, udp, Opts, Owner) ->
     case gen_udp:open(Port, [binary,
 			     {ip, IP},
 			     {active, false},
+			     {recbuf, ?UDP_RECBUF},
+			     {read_packets, ?UDP_READ_PACKETS},
 			     {reuseaddr, true}]) of
 	{ok, Socket} ->
 	    Owner ! {self(), ok},
@@ -160,36 +176,66 @@ start_listener(IP, Port, udp, Opts, Owner) ->
 	    Owner ! {self(), Err}
     end.
 
-accept(ListenSocket, Opts) ->
-    Transport = case proplists:get_bool(tls, Opts) of
-		    true -> tls;
-		    false -> tcp
-		end,
+-ifdef(USE_OLD_INET_BACKEND).
+listen(_Transport, Port, Opts) ->
+    gen_tcp:listen(Port, Opts).
+-else.
+listen(auto, Port, Opts) ->
+    gen_tcp:listen(Port, [{inet_backend, socket} | Opts]);
+listen(_Transport, Port, Opts) ->
+    gen_tcp:listen(Port, Opts).
+-endif.
+
+accept(Transport, ListenSocket, Opts) ->
+    Proxy = proplists:get_bool(proxy_protocol, Opts),
     ID = stun_logger:make_id(),
+    Opts1 = [{session_id, ID} | Opts],
     stun_logger:set_metadata(listener, Transport, ID),
     case gen_tcp:accept(ListenSocket) of
-        {ok, Socket} ->
-            case {inet:peername(Socket),
-                  inet:sockname(Socket)} of
-                {{ok, {PeerAddr, PeerPort}}, {ok, {Addr, Port}}} ->
+	{ok, Socket} when Proxy ->
+	    case p1_proxy_protocol:decode(gen_tcp, Socket, 10000) of
+		{{Addr, Port}, {PeerAddr, PeerPort}} = SP ->
+		    Opts2 = [{sock_peer_name, SP} | Opts1],
+		    ?LOG_INFO("Accepting proxied connection: ~s -> ~s",
+			      [stun_logger:encode_addr({PeerAddr, PeerPort}),
+			       stun_logger:encode_addr({Addr, Port})]),
+		    case stun:start({gen_tcp, Socket}, Opts2) of
+			{ok, Pid} ->
+			    gen_tcp:controlling_process(Socket, Pid);
+			{error, Reason} ->
+			    ?LOG_ERROR("Cannot start connection: ~s", [Reason]),
+			    gen_tcp:close(Socket)
+		    end;
+		{error, Reason} ->
+		    ?LOG_ERROR("Cannot parse proxy protocol: ~s",
+			       [inet:format_error(Reason)]),
+		    gen_tcp:close(Socket);
+		{undefined, undefined} ->
+		    ?LOG_ERROR("Cannot parse proxy protocol: unknown protocol"),
+		    gen_tcp:close(Socket)
+	    end;
+	{ok, Socket} ->
+	    case {inet:peername(Socket),
+		  inet:sockname(Socket)} of
+		{{ok, {PeerAddr, PeerPort}}, {ok, {Addr, Port}}} ->
 		    ?LOG_INFO("Accepting connection: ~s -> ~s",
 			      [stun_logger:encode_addr({PeerAddr, PeerPort}),
 			       stun_logger:encode_addr({Addr, Port})]),
-		    case stun:start({gen_tcp, Socket},
-				    [{session_id, ID}|Opts]) of
-                        {ok, Pid} ->
-                            gen_tcp:controlling_process(Socket, Pid);
-                        Err ->
-                            Err
-                    end;
-                Err ->
-                    ?LOG_ERROR("Cannot fetch peername: ~p", [Err]),
-                    Err
-            end,
-            accept(ListenSocket, Opts);
-        Err ->
-            Err
-    end.
+		    case stun:start({gen_tcp, Socket}, Opts1) of
+			{ok, Pid} ->
+			    gen_tcp:controlling_process(Socket, Pid);
+			{error, Reason} ->
+			    ?LOG_ERROR("Cannot start connection: ~s", [Reason]),
+			    gen_tcp:close(Socket)
+		    end;
+		Err ->
+		    ?LOG_ERROR("Cannot fetch peername: ~p", [Err]),
+		    gen_tcp:close(Socket)
+	    end;
+	{error, Reason} ->
+	    ?LOG_ERROR("Cannot accept connection: ~s", [Reason])
+    end,
+    accept(Transport, ListenSocket, Opts).
 
 udp_recv(Socket, Opts) ->
     case gen_udp:recv(Socket, 0) of
@@ -217,4 +263,5 @@ format_listener_error(IP, Port, Transport, Opts, Err) ->
 	       "** Transport: ~s~n"
 	       "** Options: ~p~n"
 	       "** Reason: ~p",
-	       [stun_logger:encode_addr(IP), Port, Transport, Opts, Err]).
+	       [stun_logger:encode_addr(IP), Port,
+		stun_logger:encode_transport(Transport), Opts, Err]).
