@@ -23,10 +23,11 @@
 
 -module(stun_test).
 
--export([bind_udp/2, bind_tcp/2, allocate_udp/5]).
+-export([bind_udp/2, bind_tcp/2, bind_tls/2, allocate_udp/5]).
 
 -define(STUN_IP, {127,0,0,1}).
 -define(STUN_PORT, 34780).
+-define(AUTO_PORT, 34781).
 -define(STUNS_PORT, 53490).
 -define(RECV_TIMEOUT, timer:seconds(5)).
 -define(CHANNEL, 16#4000).
@@ -63,6 +64,13 @@ add_tls_listener_test() ->
 		       ?STUN_IP, ?STUNS_PORT, tcp,
 		       [tls, {certfile, "certfile.pem"}])).
 
+-ifndef(USE_OLD_INET_BACKEND).
+add_auto_listener_test() ->
+    ?assertEqual(ok, stun_listener:add_listener(
+		       ?STUN_IP, ?AUTO_PORT, tcp,
+		       [{tls, optional}, {certfile, "certfile.pem"}])).
+-endif.
+
 bind_udp_test() ->
     TrID = mk_trid(),
     Msg = #stun{method = ?STUN_METHOD_BINDING,
@@ -102,8 +110,7 @@ bind_tls_test() ->
  		trid = TrID},
     {ok, Socket} = gen_tcp:connect(?STUN_IP, ?STUNS_PORT,
 				   [binary, {active, true}]),
-    {ok, TLSSocket} = fast_tls:tcp_to_tls(
-			Socket, [{certfile, <<"certfile.pem">>}, connect]),
+    {ok, TLSSocket} = fast_tls:tcp_to_tls(Socket, [connect]),
     ?assertEqual({ok, <<>>}, fast_tls:recv_data(TLSSocket, <<>>)),
     {ok, Addr} = fast_tls:sockname(TLSSocket),
     Pkt = stun_codec:encode(Msg),
@@ -115,12 +122,54 @@ bind_tls_test() ->
        recv(TLSSocket, <<>>, true)),
     ?assertEqual(ok, gen_tcp:close(Socket)).
 
+-ifndef(USE_OLD_INET_BACKEND).
+bind_auto_test() ->
+    %% TCP
+    TrID1 = mk_trid(),
+    Msg1 = #stun{method = ?STUN_METHOD_BINDING,
+		 class = request,
+		 trid = TrID1},
+    {ok, Socket1} = gen_tcp:connect(?STUN_IP, ?AUTO_PORT,
+				    [binary, {active, false}]),
+    {ok, Addr1} = inet:sockname(Socket1),
+    Pkt1 = stun_codec:encode(Msg1),
+    ?assertEqual(ok, gen_tcp:send(Socket1, Pkt1)),
+    ?assertMatch(
+       {ok, #stun{trid = TrID1,
+		  'XOR-MAPPED-ADDRESS' = Addr1}},
+       recv(Socket1, <<>>, false)),
+    ?assertEqual(ok, gen_tcp:close(Socket1)),
+    %% TLS
+    TrID2 = mk_trid(),
+    Msg2 = #stun{method = ?STUN_METHOD_BINDING,
+		 class = request,
+		 trid = TrID2},
+    {ok, Socket2} = gen_tcp:connect(?STUN_IP, ?AUTO_PORT,
+				    [binary, {active, true}]),
+    {ok, TLSSocket} = fast_tls:tcp_to_tls(Socket2, [connect]),
+    ?assertEqual({ok, <<>>}, fast_tls:recv_data(TLSSocket, <<>>)),
+    {ok, Addr2} = fast_tls:sockname(TLSSocket),
+    Pkt2 = stun_codec:encode(Msg2),
+    recv(TLSSocket, <<>>, true),
+    ?assertEqual(ok, fast_tls:send(TLSSocket, Pkt2)),
+    ?assertMatch(
+       {ok, #stun{trid = TrID2,
+		  'XOR-MAPPED-ADDRESS' = Addr2}},
+       recv(TLSSocket, <<>>, true)),
+    ?assertEqual(ok, gen_tcp:close(Socket2)).
+-endif.
+
 del_tcp_listener_test() ->
     ?assertEqual(ok, stun_listener:del_listener(?STUN_IP, ?STUN_PORT, tcp)).
 
 del_tls_listener_test() ->
     ?assertEqual(ok, stun_listener:del_listener(?STUN_IP, ?STUNS_PORT, tcp)).
 
+-ifndef(USE_OLD_INET_BACKEND).
+del_auto_listener_test() ->
+    ?assertEqual(ok, stun_listener:del_listener(?STUN_IP, ?AUTO_PORT, tcp)).
+-endif.
+
 allocate_udp_test() ->
     {ok, Socket} = gen_udp:open(0, [binary, {ip, ?STUN_IP}, {active, false}]),
     {ok, PeerSocket} = gen_udp:open(0, [binary, {ip, ?STUN_IP}, {active, false}]),
@@ -275,6 +324,28 @@ bind_tcp(Host, Port) ->
 	    Err
     end.
 
+bind_tls(Host, Port) ->
+    TrID = mk_trid(),
+    MsgOut = #stun{method = ?STUN_METHOD_BINDING,
+		   class = request,
+		   trid = TrID},
+    try
+	{ok, Socket} = gen_tcp:connect(Host, Port,
+				       [binary, {active, true}]),
+	{ok, TLSSocket} = fast_tls:tcp_to_tls(Socket, [connect]),
+	{ok, <<>>} = fast_tls:recv_data(TLSSocket, <<>>),
+	Pkt = stun_codec:encode(MsgOut),
+	recv(TLSSocket, <<>>, true),
+	ok = fast_tls:send(TLSSocket, Pkt),
+	{ok, MsgIn = #stun{trid = TrID,
+			   'XOR-MAPPED-ADDRESS' = _Addr}} =
+	    recv(TLSSocket, <<>>, true),
+	gen_tcp:close(Socket),
+	MsgIn
+    catch _:{badmatch, Err} ->
+	    Err
+    end.
+
 allocate_udp(Host, Port, User, Realm, Pass) ->
     try
 	{ok, Socket} = gen_udp:open(0, [binary, {active, false}]),
